/*
 * QDigiDocCommon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "QCNG.h"

#include "QPCSC.h"
#include "TokenData.h"

#include <QtCore/QDebug>
#include <QtCore/QLibrary>
#include <QtCore/QStringList.h>
#include <QtWidgets/QApplication>
#include <QtWidgets/QWidget>
#include <QtNetwork/QSslKey>

#include <qt_windows.h>
#include <WinCrypt.h>
#include <ncrypt.h>

#include <openssl/evp.h>
#include <openssl/rsa.h>

struct QCNGCache
{
	QString guid, provider, key;
	DWORD spec;
};

class QCNGPrivate
{
public:
	QCNGPrivate();

	void enumKeys( QHash<SslCertificate,QCNGCache> &cache, LPCWSTR provider, LPCWSTR scope = 0 );
	NCRYPT_KEY_HANDLE key() const;
	QByteArray prop( NCRYPT_HANDLE handle, LPCWSTR param, DWORD flags = 0 ) const;

	static int rsa_sign( int type, const unsigned char *m, unsigned int m_len,
		unsigned char *sigret, unsigned int *siglen, const RSA *rsa );

	QCNGCache selected;

	RSA_METHOD method;

	QCNG::PinStatus err;
	QHash<SslCertificate,QCNGCache> cache;
};

QCNGPrivate::QCNGPrivate()
	: err(QCNG::PinOK)
{
	method = *RSA_get_default_method();
	method.name = "QCNG";
	method.rsa_sign = rsa_sign;
}

void QCNGPrivate::enumKeys( QHash<SslCertificate,QCNGCache> &cache, LPCWSTR provider, LPCWSTR scope )
{
	QCNGCache c;
	c.provider = QString( (QChar*)provider );

	NCRYPT_PROV_HANDLE h = 0;
	SECURITY_STATUS err = NCryptOpenStorageProvider( &h, provider, 0 );

	NCryptKeyName *keyname = 0;
	PVOID pos = 0;
	while( NCryptEnumKeys( h, scope, &keyname, &pos, NCRYPT_SILENT_FLAG ) == ERROR_SUCCESS )
	{
		c.key = QString( (QChar*)keyname->pszName );
		c.spec = keyname->dwLegacyKeySpec;
		qWarning() << "key" << c.key
			<< "spec" << c.spec
			<< "alg" << QString( (QChar*)keyname->pszAlgid )
			<< "flags" << keyname->dwFlags;

		NCRYPT_KEY_HANDLE key = 0;
		err = NCryptOpenKey( h, &key, keyname->pszName, keyname->dwLegacyKeySpec, NCRYPT_SILENT_FLAG );
		NCryptFreeBuffer( keyname );
		keyname = 0;

		SslCertificate cert( prop( key, NCRYPT_CERTIFICATE_PROPERTY ), QSsl::Der );
		c.guid = prop( h, NCRYPT_SMARTCARD_GUID_PROPERTY );
		cache[cert] = c;
		NCryptFreeObject( key );
	}
	NCryptFreeObject( h );
}

NCRYPT_KEY_HANDLE QCNGPrivate::key() const
{
	NCRYPT_PROV_HANDLE prov = 0;
	NCryptOpenStorageProvider( &prov, LPCWSTR(selected.provider.utf16()), 0 );
	NCRYPT_KEY_HANDLE key = 0;
	NCryptOpenKey( prov, &key, LPWSTR(selected.key.utf16()), selected.spec, 0 );
	NCryptFreeObject( prov );
	if( qApp->activeWindow() )
	{
		WId id = qApp->activeWindow()->winId();
		NCryptSetProperty( key, NCRYPT_WINDOW_HANDLE_PROPERTY, PBYTE(&id), sizeof(id), 0 );
	}
	return key;
}

QByteArray QCNGPrivate::prop( NCRYPT_HANDLE handle, LPCWSTR param, DWORD flags ) const
{
	if( !handle )
		return QByteArray();
	DWORD size = 0;
	if( NCryptGetProperty( handle, param, 0, 0, &size, flags ) )
		return QByteArray();
	QByteArray data( size, 0 );
	if( NCryptGetProperty( handle, param, PBYTE(data.data()), size, &size, flags ) )
		return QByteArray();
	return data;
}

int QCNGPrivate::rsa_sign( int type, const unsigned char *m, unsigned int m_len,
	unsigned char *sigret, unsigned int *siglen, const RSA *rsa )
{
	if( type != NID_md5_sha1 && m_len != 36 ) //ssl
		return 0;
	QCNG *d = (QCNG*)RSA_get_app_data( rsa );
	QByteArray sig = d->sign( type, QByteArray( (char*)m, m_len ) );
	if( sig.isEmpty() )
		return 0;
	*siglen = sig.size();
	std::memcpy( sigret, sig.constData(), sig.size() );
	return 1;
}



QCNG::QCNG( QObject *parent )
:	QObject( parent )
,	d( new QCNGPrivate )
{
}

QCNG::~QCNG()
{
	delete d;
}

QByteArray QCNG::decrypt( const QByteArray &data )
{
	d->err = PinUnknown;
	DWORD size = 256;
	QByteArray res( size, 0 );
	NCRYPT_KEY_HANDLE k = d->key();
	DWORD err = NCryptDecrypt( k, PBYTE(data.constData()), data.size(), 0,
		PBYTE(res.data()), res.size(), &size, NCRYPT_PAD_PKCS1_FLAG );
	NCryptFreeObject( k );
	switch( err )
	{
	case ERROR_SUCCESS:
		d->err = PinOK;
		res.resize( size );
		return res;
	case SCARD_W_CANCELLED_BY_USER:
		d->err = PinCanceled; break;
	default: break;
	}
	return QByteArray();
}

QCNG::PinStatus QCNG::lastError() const { return d->err; }

QCNG::Certs QCNG::certs() const
{
	qWarning() << "Start enumerationg providers";
	QHash<SslCertificate,QCNGCache> cache;
	DWORD count = 0;
	NCryptProviderName *names = 0;
	NCryptEnumStorageProviders( &count, &names, NCRYPT_SILENT_FLAG );
	for( DWORD i = 0; i < count; ++i )
	{
		qWarning() << "Found provider" << QString( (QChar*)names[i].pszName );
		if( wcscmp( names[i].pszName, MS_SMART_CARD_KEY_STORAGE_PROVIDER ) == 0 )
		{
			foreach( const QString &reader, QPCSC().readers() )
			{
				qWarning() << reader;
				QString scope = QString( "\\\\.\\%1\\" ).arg( reader );
				d->enumKeys( cache, names[i].pszName, LPCWSTR(scope.utf16()) );
			}
		}
		else
			d->enumKeys( cache, names[i].pszName );
	}
	NCryptFreeBuffer( names );
	d->cache = cache;
	qWarning() << "End enumerationg providers";

	Certs result;
	for(QHash<SslCertificate,QCNGCache>::const_iterator i = cache.constBegin(); i != cache.constEnd(); ++i)
	{
		result[i.key()] = i.key().type() & SslCertificate::EstEidType || i.key().type() & SslCertificate::DigiIDType ?
			i.value().guid : i.key().subjectInfo( QSslCertificate::CommonName );
	}
	return result;
}

Qt::HANDLE QCNG::key()
{
	NCRYPT_KEY_HANDLE k = d->key();
	QSslCertificate cert( d->prop( k, NCRYPT_CERTIFICATE_PROPERTY ), QSsl::Der );
	NCryptFreeObject( k );
	RSA *rsa = RSAPublicKey_dup( (RSA*)cert.publicKey().handle() );
	if ( !rsa )
		return 0;

	RSA_set_method( rsa, &(d->method) );
	rsa->flags |= RSA_FLAG_SIGN_VER;
	RSA_set_app_data( rsa, this );
	EVP_PKEY *key = EVP_PKEY_new();
	EVP_PKEY_set1_RSA( key, rsa );
	RSA_free( rsa );
	return Qt::HANDLE(key);
}

QStringList QCNG::readers() const
{
	return QPCSC().drivers();
}

TokenData QCNG::selectCert( const SslCertificate &cert )
{
	qWarning() << "Select:" << cert.subjectInfo( "CN" );
	if( !d->cache.contains( cert ) )
		return TokenData();

	d->selected = d->cache[cert];
	qWarning() << "Found:" << d->selected.guid << d->selected.key;
	TokenData t;
	t.setCard( cert.type() & SslCertificate::EstEidType || cert.type() & SslCertificate::DigiIDType ?
		d->selected.guid : cert.subjectInfo( QSslCertificate::CommonName ) );
	t.setCert( cert );

	return t;
}

QByteArray QCNG::sign( int method, const QByteArray &digest ) const
{
	d->err = PinUnknown;
	BCRYPT_PKCS1_PADDING_INFO padInfo;
	padInfo.pszAlgId = 0;
	switch( method )
	{
	case NID_sha1: padInfo.pszAlgId = NCRYPT_SHA1_ALGORITHM; break;
	case NID_sha224: padInfo.pszAlgId = L"SHA224"; break;
	case NID_sha256: padInfo.pszAlgId = NCRYPT_SHA256_ALGORITHM; break;
	case NID_sha384: padInfo.pszAlgId = NCRYPT_SHA384_ALGORITHM; break;
	case NID_sha512: padInfo.pszAlgId = NCRYPT_SHA512_ALGORITHM; break;
	case NID_md5_sha1: //padInfo.pszAlgId = L"SHAMD5"; break;
	default: break;
	}

	DWORD size = 256;
	QByteArray res( size, 0 );
	NCRYPT_KEY_HANDLE k = d->key();
	DWORD err = NCryptSignHash( k, &padInfo, PBYTE(digest.constData()), digest.size(),
		PBYTE(res.data()), res.size(), &size, BCRYPT_PAD_PKCS1 );
	NCryptFreeObject( k );
	switch( err )
	{
	case ERROR_SUCCESS:
		d->err = PinOK;
		res.resize( size );
		return res;
	case SCARD_W_CANCELLED_BY_USER:
		d->err = PinCanceled; break;
	default: break;
	}
	return QByteArray();
}
